""" 
This module was created to realize
Statistical Tests in order to quantify
distributions relations.

"""


from typing import Tuple, List
from scipy.stats import ttest_ind as stats_ttest_ind
from scipy.stats import chi2_contingency as stats_chi2_contingency
from scipy.stats import mannwhitneyu as stats_mannwhitneyu
from pandas import DataFrame, crosstab


def mann_whitney_u_test(
    dataframe: DataFrame,
    numerical_variable: str,
    categorical_variable: str,
) -> Tuple[float, float]:
    """Perform the Mann-Whitney U rank test between 1 numerical and 1 categorical column 
    with only two distinct classes..
    
    Perform the Mann-Whitney U rank test on two independent samples. The 
    Mann-Whitney U test is a nonparametric test of the 
    null hypothesis that the distribution underlying sample 
    x is the same as the distribution underlying 
    sample y. It is often used as a test of difference 
    in location between distributions.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame
    numerical_variable : str
        Name of numerical column for the Mann-Whitney U test.
    categorical_variable : str
        Name of categorical column for the Mann-Whitney U test.

    Returns
    -------
    Tuple[float, float]
        The values of Test Statistic and the p-value.
    """
    
    df = dataframe.loc[:, [numerical_variable, categorical_variable]]
    df = df.dropna()    
    categorical_values = list(
        df[categorical_variable] \
        .value_counts() \
        .index
    )
    
    assert len(categorical_values) == 2, \
        "Categorical column must have only 2 distinct classes."
        
    statistic, p_value = stats_ttest_ind(
        df[
            df[categorical_variable] == categorical_values[0]
        ][numerical_variable], 
        df[
            df[categorical_variable] == categorical_values[1]
        ][numerical_variable]
    )
    statistic, p_value = stats_mannwhitneyu(
        df[numerical_variable],
        df[categorical_variable],
    )
    return statistic, p_value


def t_test_ind(
    dataframe: DataFrame,
    numerical_variable: str,
    categorical_variable: str,
) -> Tuple[float, float]:
    """Perform the Two-Sample T-Test between 1 numerical and 1 categorical column 
    with only two distinct classes.
    
    Calculate the T-test for the means of two independent samples of scores.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame
    numerical_variable : str
        Name of numerical column for the T-Test
    categorical_variable : str
        Name of categorical column for the T-Test

    Returns
    -------
    Tuple[float, float]
        The values of Test Statistic and the p-value.
    """
    temp_df = dataframe[[numerical_variable, categorical_variable]]
    temp_df = temp_df.dropna()    
    categorical_values = list(
        temp_df[categorical_variable] \
        .value_counts() \
        .index
    )
    
    assert len(categorical_values) == 2, \
        "Categorical column must have only 2 distinct classes."
        
    t_stat, p = stats_ttest_ind(
        temp_df[
            temp_df[categorical_variable] == categorical_values[0]
        ][numerical_variable], 
        temp_df[
            temp_df[categorical_variable] == categorical_values[1]
        ][numerical_variable]
    )
    return t_stat, p


def chi2_contingency(
    dataframe: DataFrame,
    categorical_column1: str,
    categorical_column2: str
) -> Tuple[float, float, int, List[List[float]]]:
    """Realize Chi-Squared contigency test between 2 categorical 
    columns of DataFrame.
    
    Chi-square test of independence of variables in a contingency table.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame
    categorical_column1 : str
        Name of first categorical column
    categorical_column2 : str
        Name of second categorical column

    Returns
    -------
    Tuple[float, float, int, List[List[float]]]
        The Chi-squared Stat value, the p-value, the degrees of freedom
        and the expected matrix.
    """
    
    df = dataframe.copy()
    all_categories_col1 = list(df[categorical_column1].value_counts().index)
    all_categories_col2 = list(df[categorical_column2].value_counts().index)
    cross_tab = crosstab(df[categorical_column1], df[categorical_column2], margins=True)
    observed = cross_tab.iloc[0:-1, 0:-1]
    observed.columns = all_categories_col2
    observed.index = all_categories_col1
    chi_squared_stat, p_value, degree_of_freedom, expected \
        = stats_chi2_contingency(observed=observed)
    return chi_squared_stat, p_value, degree_of_freedom, expected


""" 
This module was created to plot useful graphics
which will be mostly used in EDA.

"""


from drtools.utils import list_ops, to_title
from drtools.file_manager import create_directories_of_path
from drtools.data_science.data_handle import prepare_bins, binning_numerical_variable
from types import FunctionType
from typing import Dict, List, Tuple, Union
import pandas as pd
from typing import Any
from pandas import DataFrame, Series
from numpy import inf
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
from sklearn import metrics
import math
from sklearn.inspection import permutation_importance as sklearn_permutation_importance
# import shap


COLORS = [
    {'PRIMITIVE': '#0D4C92', 'PRETTY': '#0D4C92'}, 
    {'PRIMITIVE': '#59C1BD', 'PRETTY': '#59C1BD'}, 
    {'PRIMITIVE': '#A0E4CB', 'PRETTY': '#A0E4CB'}, 
    {'PRIMITIVE': '#CFF5E7', 'PRETTY': '#CFF5E7'}, 
    
    {'PRIMITIVE': 'blue', 'PRETTY': 'cornflowerblue'}, 
    {'PRIMITIVE': 'red', 'PRETTY': 'indianred'}, 
    {'PRIMITIVE': 'green', 'PRETTY': 'springgreen'}, 
    {'PRIMITIVE': 'purple', 'PRETTY': 'magenta'}, 
]


def na_heatmap(
    dataframe: DataFrame,
    isna=True,
    figsize=(10, 5),
    vertical_sep: bool=False,
    rotate_x: int=45,
    nullity: bool=True,
    sort_columns: bool=True,
) -> None:    
    """Plot na distribution among columns of DataFrame. 

    Parameters
    ----------
    df : DataFrame
        DataFrame will be ploted
    isna : bool, optional
        If True, plot will show distribution of
        columns where na, else plot will show
        distribution of columns where not na, by default True
    figsize : Tuple[int, int], optional
        Size of graph, by default (10, 5)
    vertical_sep : bool, optional
        If True, will separate columns with a white line, 
        by default False.
    rotate_x : int, optional
        Value to rotate x labels, by default 45.    
    nullity : bool, optional
        If True, will display the NA percentage by column, 
        by default True.
    sort_columns : bool, optional
        If True, the columns of DataFrame will be sorted, 
        by default True.
        
    Returns
    ------
        None
            **None** is returned
    """
    
    operation = {
        True: 'isna',
        False: 'notna',
    }
    
    fig, ax = plt.subplots(figsize=figsize)
    
    df = dataframe.copy()
    
    if sort_columns:
        df = df.reindex(sorted(df.columns), axis=1)
    
    sns.heatmap(
        getattr(df, operation[isna])(),
        ax=ax
    )
        
    if vertical_sep:
        for i in range(df.shape[1] + 1):
            ax.axvline(i, color='white', lw=2)
        
    if rotate_x > 0:
        fig.autofmt_xdate(rotation=rotate_x)
        
    if nullity:
        ax2 = ax.twiny()
        ax2.set_xlim(ax.get_xlim())
        ax2.set_xticks([i + 0.5 for i in range(df.shape[1])])        
        tick_labels = []
        for col in df.columns:
            
            percent = getattr(df[col], operation[isna])().sum() * 100 / df.shape[0]
            if 99 < percent < 100:
                percent = math.floor(percent)
            else:
                percent = math.ceil(percent)
                
            tick_labels.append(f'{percent}%')
        ax2.set_xticklabels(tick_labels)


def savefig(
    save_path: str
) -> None:
    """Save figure of subplot in selected path

    Parameters
    ----------
    save_path : str
        Path to save figure
        
    Returns
    ------
        None
            **None** is returned
    """
    plt.tight_layout()
    plt.savefig(save_path)


class Legend:
    """Class to handle the legends on plots
    
    """
    
    def __init__(
        self,
        active: bool=False,
        title: str='',
        labels: List[str]=[],
        loc='upper right',
        bbox_to_anchor: Tuple[float, float]=None,
    ) -> None:
        args = locals().copy()
        args = {k: v for k, v in args.items() if k != 'self'}
        for k, v in args.items():
            setattr(self, k, v)
            
    def prepare(self, subplot: plt) -> None:
        if not self.active:
            return None
        subplot.legend(
            title=self.title, 
            loc=self.loc, 
            labels=self.labels,
            bbox_to_anchor=self.bbox_to_anchor
        )
        return None


class TwinxValuesCalculation:
    
    def __init__(
        self,
        column: str=None,
        apply: FunctionType=None,
        agg: str=None,
        filter_function: FunctionType=None,
    ) -> None:
        self.column = column
        self.apply = apply
        self.agg = agg
        self.filter_function = filter_function


class Twinx:
    """Class to handle Twinx axe on the plots
    
    """
    
    def __init__(
        self,  
        active: bool=False,
        x: Union[List, Series]=[],
        y: Union[List, Series]=[],
        ylabel: str='Twinx Axe',
        ylim: Tuple[float, float]=(0, 1),
        color: str='tab:olive',
        label: str='',
        marker: str='o',
        grid: str=None,
        fillna: any=0,        
        legend: Legend=Legend(),
        twinx_values_calculation: TwinxValuesCalculation=TwinxValuesCalculation(),
        apply: List[FunctionType]=None,
    ) -> None:
        # if active \
        # and twinx_values_calculation.column is None:
        #     raise Exception(
        #         'If "active" is True, "twinx_values_calculation.column" need to be specified.'
        #     )  
        # if active \
        # and (
        #     twinx_values_calculation.apply is None
        #     and twinx_values_calculation.agg is None
        # ):
        #     text = 'If "active" is True, "twinx_values_calculation.apply" '
        #     text += 'or "twinx_values_calculation.agg" need to be specified.'
        #     raise Exception(text)  
             
        args = locals().copy()
        args = {k: v for k, v in args.items() if k != 'self'}
        for k, v in args.items():
            setattr(self, k, v)
            
    def prepare(self, subplot: plt) -> None:
        if not self.active:
            return None
        if type(self.x) == Series:
            self.x = self.x.values
        if type(self.y) == Series:
            self.y = self.y.values
        twinx = subplot.twinx()
        twinx.plot(
            self.x, 
            self.y, 
            color=self.color, 
            label=self.label, 
            marker=self.marker
        )
        if self.ylabel is not None:
            twinx.set_ylabel(self.ylabel)
        if self.ylim is not None:
            twinx.set_ylim(*self.ylim)
        twinx.grid(self.grid)        
        
        # twinx.set_yticks(
        #     np.linspace(
        #         twinx.get_yticks()[0], 
        #         twinx.get_yticks()[-1], 
        #         len(subplot.get_yticks())
        #     )
        # )
        
        self.legend.prepare(twinx)
        
        if self.apply is not None:
            for func in self.apply:
                func(twinx, subplot)
        
        return None


def bar(
    dataframe: DataFrame,
    analysis_variable: str,
    hue: str=None,
    figsize: Tuple[int, int]=(10, 5),
    fillna: float=0,
    na_as_class: bool=False,
    ignore_columns: List=[],
    sort_index_attr: dict={
        'ascending': True
    },
    sort_values_attr: dict={},
    rotate_x: int=0,
    twinx: Union[Twinx, List[Twinx]]=Twinx(),
    legend: Legend=Legend(
        active=True, loc='upper right'
    ),
    savefig_path: str='',
) -> None:
    """Plot custom stacked bar.

    Parameters
    ----------
    df : DataFrame
        DataFrame containing both analysis variable and hue variable
    analysis_variable : str
        Variable to be ploted
    hue : str, optional
        Hue variable, by default None.
    figsize : Tuple[int, int], optional
        Size of graph, by default (10, 5)
    fillna : float, optional
        Value to fill NA values of DataFrame, by default 0
    na_as_class : bool, optional
        If True, NA values will be transformod in a class, 
        if False, NA values will not be considered when 
        ploting, by default False
    ignore_columns : List, optional
        Ignore these columns when plot data. These columns
        will not be considered to plot data, by default []
    sort_index_attr : dict, optional
        Attributes will be applied on sorting
        index, by default { 'ascending': True }
    sort_values_attr : dict, optional
        Attributes will be applied on sorting
        values, by default { }
    savefig_path : str, optional
        Path to save figure, by ''
    rotate_x : int, optional
        Value to rotate x labels, by default 0.    
    twinx : Union[Twinx, List[Twinx]], optional
        Attributes for the twinx axe, by default Twinx()
    legend : Legend, optional
        Attributes for the legend, by default Legend()
        
    Returns
    -------
    None
        **None** is returned
    """
    
    sns.set_style('darkgrid')
    df = dataframe.copy()
    
    if not hue:
        hue = '__temp_fill__'
        df[hue] = hue
        
    if na_as_class:
        df[analysis_variable] = df[analysis_variable].fillna('nan')
        
    margins_name = '__total__'
    
    df = pd.crosstab(
        df[analysis_variable], 
        df[hue],
        margins=True,
        margins_name=margins_name
    )
    df = df.fillna(fillna)
    
    work_columns = list_ops(df.columns, ignore_columns + [margins_name])

    fig, ax1 = plt.subplots(figsize=figsize)

    # Define Labels

    ax1.set_xlabel(analysis_variable)
    ax1.set_ylabel('Count')
    
    df.index = df.index.astype(str)
    df = df.sort_index(**sort_index_attr)
    if sort_values_attr:
        df = df.sort_values(**sort_values_attr)
    df = df.drop(margins_name, axis='index')
    df = df.drop(margins_name, axis='columns')
        
    df.index = df.index.map(str)

    bottom = None
    for idx, col in enumerate(work_columns):
        color = COLORS[idx % len(COLORS)]['PRETTY']
        if idx == 0:
            ax1.bar(
                df.index, 
                df[col], 
                color=color,
            )
            bottom = df[col]
        else:                    
            ax1.bar(
                df.index, 
                df[col], 
                bottom=bottom, 
                color=color
            )
            
            bottom = bottom + df[col]

    if hue != '__temp_fill__':
        legend.title = hue
        legend.labels = [to_title(str(col)) for col in work_columns]
        legend.prepare(ax1)
        
    # Twin Axes
    def plot_twinx(
        twinx_obj: Twinx, 
        subplot,
    ):
        if twinx_obj.active:
            
            index_len = len(df.index)
            twinx_obj.x = np.linspace(0, index_len - 1, index_len)
            
            twinx_values_calculation = twinx_obj.twinx_values_calculation
            values_from_column = twinx_values_calculation.column or hue
            apply_operation = twinx_values_calculation.apply
            agg_operation = twinx_values_calculation.agg
            filter_op = twinx_values_calculation.filter_function
            
            temp_df = dataframe.copy()
            
            if na_as_class:
                temp_df[analysis_variable] = temp_df[analysis_variable].fillna('nan')
                
            if filter_op is not None:
                twinx_obj.y = filter_op(temp_df.copy())
            else:
                twinx_obj.y = temp_df.copy()
                
            twinx_obj.y = twinx_obj.y.groupby([analysis_variable])
            twinx_obj.y = twinx_obj.y[values_from_column]
            if apply_operation is not None:
                twinx_obj.y = twinx_obj.y.apply(apply_operation)
            else:
                twinx_obj.y = twinx_obj.y.agg(agg_operation)
            twinx_obj.y = twinx_obj.y.reindex(index=list(df.index))
            twinx_obj.y = twinx_obj.y.values
            
            twinx_obj.prepare(subplot)
    
    if type(twinx) != list:
        twinx = [twinx]
        
    # if type(twinx) == list:
    for idx, curr_twinx in enumerate(twinx):
        if idx == 0:
            plot_twinx(twinx_obj=curr_twinx, subplot=ax1)
        else:
            color = COLORS[idx % len(COLORS)]['PRETTY']
            curr_twinx.color = color
            curr_twinx.ylabel = None
            curr_twinx.apply = [
                lambda twinx_plt, subplot: twinx_plt.set_yticklabels([])
            ]
            plot_twinx(twinx_obj=curr_twinx, subplot=ax1)
    # else:
    #     plot_twinx(
    #         twinx_obj=twinx, 
    #         subplot=ax1,                   
    #     )

    # Display
    plt.title(f'Bar plot of {analysis_variable}')
    
    if rotate_x > 0:
        fig.autofmt_xdate(rotation=rotate_x)
    
    if savefig_path:
        savefig(savefig_path)
        
    plt.show()


def hist(
    dataframe: DataFrame,
    analysis_variable: str,
    hue: str=None,
    analysis_range: Tuple[float, float]=(-inf, inf),
    figsize: Tuple[int, int]=(10, 5),
    bins: Union[int, List[float]]=20,
    kde: bool=False,
    extra_information: bool=True,
    alpha: float=0.4,
    mean: bool=False,
    twinx: Union[Twinx, List[Twinx]]=Twinx(fillna=0),
    legend: Legend=Legend(
        active=True, loc='upper right'
    ),
    savefig_path: str='',
) -> None:
    """Plot histogram with hue on binary variable

    Parameters
    ----------
    df : DataFrame
        DataFrame containing both analysis variable and hue variable
    analysis_variable : str
        Variable to be ploted
    hue : str
        Hue variable
    analysis_range : Tuple[float, float], optional
        Range of limits in which histogram
        will be plotted. Do not include upper range limit
        , by default (-inf, inf)
    figsize : Tuple[int, int], optional
        Size of ploted graph, by default (10, 5)
    bins : Union[int, List[float]], optional
        Number of bins or interval of values
        to binning ``analysis_variable``, by default 100
    kde : bool, optional
        Plot density, by default False
    extra_information : bool, optional
        Show extra information about ploted data, 
        by default True
    alpha : float, optional
        Alpha value of bins on hist plot. Must be between 0 and 1, by default 0.4
    mean : bool, optional
        Show mean on graph, by default False
    twinx : Union[Twinx, List[Twinx]], optional
        Attributes for the twinx axe, by default Twinx(fillna=0)
    legend : Legend, optional
        Attributes for the legend, by default Legend( active=True, loc='upper right' )
    savefig_path : str, optional
        Path to save figure, by ''   
        
    Returns
    ------
    None
        **None** is returned    
    """
    
    sns.set_style('darkgrid')
    df = dataframe.copy()
    original_total_points = len(df)
    
    temp_analysis_range = [None, None]
    temp_analysis_range[0] = max(bins[0], analysis_range[0]) \
        if hasattr(bins, '__iter__') \
        else analysis_range[0]
    temp_analysis_range[1] = min(bins[-1], analysis_range[1]) \
        if hasattr(bins, '__iter__') \
        else analysis_range[1]
    analysis_range = tuple(temp_analysis_range)
    lower_boundary = analysis_range[0]
    upper_boundary = analysis_range[1]
    
    try:
        v = df[analysis_variable].ge(lower_boundary) \
            & df[analysis_variable].lt(upper_boundary)
        df = df[v]
    except Exception as exc:
        print(f'Error when apply filters: {exc}')
        df = df.copy()        

    
    if not hue:
        hue = '__temp_fill__'
        df[hue] = hue
        
    v = df[hue].value_counts()
    series_data = {}
    for val in v.index:
        series_data[val] = df[df[hue] == val][analysis_variable]

    fig, ax1 = plt.subplots(figsize=figsize)
    
    # ax1.set_xlim([lower_boundary, upper_boundary])
    
    def mean_constructor(
        plt_instance,
        data_series: Series,
        mean_attr: Dict,
        annotate_position: int,
    ) -> None:
        plt_instance.vlines(
            data_series.mean(),
            **mean_attr
        )
        plt_instance.annotate(
            f'mean: {data_series.mean(): .2f}',
            xy=(data_series.mean() * 1.02, annotate_position)
        )
    
    # set bins
    sorted_values = df[analysis_variable] \
        .dropna() \
        .sort_values() \
        .values
    bins_interval, points, labels \
        = prepare_bins(
        bins,
        smallest_value=lower_boundary if lower_boundary != -inf else sorted_values[0], 
        bigger_value=upper_boundary if upper_boundary != inf else sorted_values[-1],
        include_upper=upper_boundary == inf
    )

    total_points = 0
    legend_labels = []
    for index, key in enumerate(series_data): 
        curr_series_data = series_data[key]
        curr_series_sorted = curr_series_data.sort_values()
        non_correspondences = pd.cut(
            curr_series_sorted,
            bins=bins_interval,
            right=False
        ).isna().sum()
        if non_correspondences != 0:
            total_points = total_points + len(
                curr_series_sorted.reset_index(drop=True)[:-non_correspondences]
            )
        else:
            total_points = total_points + len(
                curr_series_sorted
            )
        if hue != '__temp_fill__':
            legend_labels.append(f'{hue} is {str(key)}')
        sns.histplot(
            curr_series_data,
            color=COLORS[index % len(COLORS)]['PRETTY'],
            kde=kde,
            bins=bins_interval,
            alpha=alpha,
            ax=ax1
        )
        
    if hue != '__temp_fill__':
        legend.labels = legend_labels
        legend.prepare(ax1)

    if mean:
        max_value = -inf
        for index, key in enumerate(series_data): 
            curr_series_data = series_data[key]
            curr_max = pd.cut(
                curr_series_data.sort_values(),
                bins=bins_interval,
                right=False,
            ).value_counts().sort_values().values[-1]
            max_value = max(
                max_value, 
                curr_max
            )
        for index, key in enumerate(series_data):
            ser_data = series_data[key]
            real_max_value = max_value * 1.02
            mean_constructor(
                ax1,
                ser_data,
                {
                    'ymin': 0,
                    'ymax': real_max_value,
                    'color': COLORS[index % len(COLORS)]['PRIMITIVE']
                },
                real_max_value,
            )            
            
    # Twin Axes
    def plot_twinx(twinx_obj: Twinx, subplot):
        if twinx_obj.active:
            
            temp_df = binning_numerical_variable(
                df,
                analysis_variable,
                bins=bins_interval,
                binning_column_prefix='binning'
            )
            
            twinx_obj.x = points
            
            if len(twinx_obj.y) == 0:                
                twinx_values_calculation = twinx_obj.twinx_values_calculation
                values_from_column = twinx_values_calculation.column or hue
                apply_operation = twinx_values_calculation.apply
                agg_operation = twinx_values_calculation.agg
                filter_op = twinx_values_calculation.filter_function
                
                if filter_op is not None:
                    twinx_obj.y = filter_op(temp_df.copy())
                else:
                    twinx_obj.y = temp_df.copy()
                
                twinx_obj.y = twinx_obj.y.groupby([f'binning_{analysis_variable}'])
                twinx_obj.y = twinx_obj.y[values_from_column]
                if apply_operation is not None:
                    twinx_obj.y = twinx_obj.y.apply(apply_operation)
                else:
                    twinx_obj.y = twinx_obj.y.agg(agg_operation)
                twinx_obj.y = twinx_obj.y.values
            
            else:
                pass
            
            twinx_obj.prepare(subplot)
            
        
    if type(twinx) != list:
        twinx = [twinx]
        
    # if type(twinx) == list:
    for idx, curr_twinx in enumerate(twinx):
        if idx == 0:
            plot_twinx(twinx_obj=curr_twinx, subplot=ax1)
        else:
            color = COLORS[(idx-1) % len(COLORS)]['PRETTY']
            if curr_twinx.color == 'tab:olive':
                curr_twinx.color = color
            curr_twinx.ylabel = None
            curr_twinx.apply = [
                lambda twinx_plt, subplot: twinx_plt.set_yticklabels([])
            ]
            plot_twinx(twinx_obj=curr_twinx, subplot=ax1)
    # else:
    #     plot_twinx(
    #         twinx_obj=twinx, 
    #         subplot=ax1,                   
    #     )

    
    extra_information_text = ''
    if extra_information:
        trace_text = '-' * 100
        extra_information_text += f'{trace_text}'
        extra_information_text += '\n'
        extra_information_text += f'Information of {analysis_variable} variable'
        extra_information_text += '\n'
        extra_information_text += '\n'
        na_points = dataframe[analysis_variable].isna().sum()
        non_na_points = dataframe[analysis_variable].notna().sum()
        extra_information_text += f'<NA> points: '
        extra_information_text += f'{na_points:,} | '
        extra_information_text += f'{round(na_points * 100 / original_total_points, 3)}% '
        extra_information_text += f'#{original_total_points:,}'
        extra_information_text += '\n'
        extra_information_text += f'Non <NA> points: '
        extra_information_text += f'{non_na_points:,} | '
        extra_information_text += f'{round(non_na_points * 100 / original_total_points, 3)}% '
        extra_information_text += f'#{original_total_points:,}'
        extra_information_text += '\n'
        extra_information_text += f'Total Points on range '
        extra_information_text += f'[{analysis_range[0]:,}, {analysis_range[1]:,}): '
        extra_information_text += f'{total_points:,} | '
        extra_information_text += f'{round(total_points * 100 / non_na_points, 3)}% '
        extra_information_text += f'#{non_na_points:,}'
        if hue != '__temp_fill__':
            extra_information_text += '\n'
            for key in series_data:
                ser_data = series_data[key]
                text_value = str(key)
                curr_len = len(ser_data)
                percentage = round(curr_len * 100 / total_points, 2)
                extra_information_text += '\n'
                extra_information_text += f'- {hue} = {text_value} | {percentage}% #{curr_len:,}'
        extra_information_text += '\n'
        extra_information_text += f'{trace_text}'
        extra_information_text += '\n'
        extra_information_text += '\n'
    
    title_text = extra_information_text + f'Histogram of {analysis_variable}'
    
    ax1.set_title(title_text)
    ax1.set_xlabel(analysis_variable)
    
    if savefig_path:
        savefig(savefig_path)
        
    plt.show()


def roc_curve(
    y_true: List[float],
    y_score: List[float],
    pos_label: Union[int,str]=None,
    figsize: Tuple[int, int]=(5, 5),
    savefig_path: str='',
    **kwargs
) -> None:
    """Generate ROC Curve with AUC value.csv

    Parameters
    ----------
    y_true : List[float]
        True binary labels
    y_score : List[float]
        Target scores, can either be probability 
        estimates of the positive class, confidence 
        values, or non-thresholded measure of 
        decisions (as returned by “decision_function” 
        on some classifiers).
    pos_label : int, optional
        The label of the positive class. When 
        pos_label=None, if y_true is in {-1, 1} 
        or {0, 1}, pos_label is set to 1, 
        otherwise an error will be raised, 
        by default 1
    figsize : Tuple[int, int], optional
        Size of graph, by default (5, 5)
    savefig_path : str, optional
        Path to save figure, by ''
    kwargs : any, optional
        Extra arguments that will be passed to the 
        function metrics.roc_curve.
    """
    
    fpr, tpr, thresholds = metrics.roc_curve(
        y_true=y_true, 
        y_score=y_score, 
        pos_label=pos_label,
        **kwargs
    )
    roc_auc = metrics.auc(fpr, tpr)
    fig, ax1 = plt.subplots(figsize=figsize)
    ax1.set_title('ROC Curve')
    ax1.plot(fpr, tpr, 'b', label='Model (AUC = %0.2f)' % roc_auc)
    ax1.plot([0, 1], [0, 1], 'r', label='Random')
    ax1.legend(loc='lower right')
    ax1.set_xlim([0, 1])
    ax1.set_ylim([0, 1])
    ax1.set_ylabel('sensitivity')
    ax1.set_xlabel('1-specificity')
    
    if savefig_path:
        savefig(savefig_path)
    
    plt.show()
    
    
def lightgbm_feature_importance(
    model, 
    variables: List[str], 
    num=20, 
    figsize=(40, 20), 
    savefig_path: str=None
) -> None:
    """Plot importance from Light GBM model

    Parameters
    ----------
    model : Any
        The light gbm model instance
    variables : List[str]
        List of variables to verify importance.
    num : int, optional
        Num of variables to plot, by default 20
    figsize : Tuple[int, int], optional
        Size of graph, by default (40, 20)
    savefig_path : str, optional
        Path to save figure, by None
    """
    feature_imp = pd.DataFrame({
        'Value': model.feature_importance(),
        'Feature': variables
    })
    
    fig, ax = plt.subplots(figsize=figsize)
    
    # sns.set(font_scale=5)
    sns.barplot(
        x="Value", 
        y="Feature", 
        data=feature_imp.sort_values(by="Value", ascending=False)[0:num],
        ax=ax
    )
    ax.set_title('LightGBM Features (avg over folds)')
    
    if savefig_path is not None:
        create_directories_of_path(savefig_path)
        savefig(savefig_path)
    plt.show()
    
    
def permutation_importance(
    model: Any,
    X: DataFrame,
    y: Series,
    n_repeats: int=3,
    random_state: int=42,
    n_jobs: int=2,
    figsize=(40, 20), 
    savefig_path: str=None
) -> None:
    """Plot permutation importance

    Parameters
    ----------
    model : Any
        Model instance
    X : DataFrame
        X data as DataFrame
    y : Series
        y data as Series
    n_repeats : int, optional
        Number of times to permute a feature, by default 3
    random_state : int, optional
        Pseudo-random number generator to control the 
        permutations of each feature. Pass an int to 
        get reproducible results across function calls, by default 42
    n_jobs : int, optional
        Number of jobs to run in parallel, by default 2
    figsize : Tuple[int, int], optional
        Size of graph, by default (40, 20)
    savefig_path : str, optional
        Path to save figure, by None
    """
    
    # model_copy = deepcopy(model)
    
    # if 'predict' in dir(model_copy):
    #     setattr(model_copy, 'fit', model_copy.predict)
    
    result = sklearn_permutation_importance(
        model,
        X,
        y,
        n_repeats=n_repeats,
        random_state=random_state,
        n_jobs=n_jobs,
    )
    
    sorted_importances_idx = result.importances_mean.argsort()
    importances = DataFrame(
        result.importances[sorted_importances_idx].T,
        columns=X.columns[sorted_importances_idx]
    )
    
    ax = importances.plot.box(vert=False, whis=10)
    ax.set_title("Permutation Importances")
    ax.axvline(x=0, color='k', linestyle='--')
    ax.figure.tight_layout()
    
    if savefig_path is not None:
        create_directories_of_path(savefig_path)
        savefig(savefig_path)
        
    plt.show()
    

# def shap_importance(
#     model: Any,
#     X: DataFrame,
# ) -> None:
#     """Plot most important features using Shap Importance

#     Parameters
#     ----------
#     model : Any
#         Model instance
#     X : DataFrame
#         X data as DataFrame
#     """
#     # Explain model predictions using shap library:
#     explainer = shap.TreeExplainer(model)
#     shap_values = explainer.shap_values(X)
    
#     # Plot summary_plot
#     shap.summary_plot(shap_values, X, plot_type='bar')
    
    
    
    
    
    


""" 
This module was created to handle
interaction on Web Site page.

"""


from typing import List
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as ExpectedConditions
from selenium.common.exceptions import TimeoutException
import logging
from selenium.webdriver.chrome.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement


def wait_for_element(
    driver: WebDriver, 
    expected_conditions: ExpectedConditions,
    delay: int=5
) -> WebDriver:
    """Wait for element until expected conditions were satisfied or delay time.

    Parameters
    ----------
    driver : WebDriver
        The Selenium WebDriver
    expected_conditions : FunctionType
        Desired expected conditions to be satisfied
    delay : int, optional
        Max waiting time in seconds until conditions 
        were satisfied, by default 5

    Returns
    -------
    WebDriver
        The WebDriver

    Raises
    ------
    TimeoutException
        If conditions were not satisfied whithin delay time.
    """
    
    my_elem = None
    try:
        my_elem = WebDriverWait(driver, delay).until(expected_conditions)
        logging.debug("Expected conditions were satisfied!")
    except TimeoutException:
        raise TimeoutException("Expected conditions took too much time!")
    return my_elem


def close_current_tab(
    driver: WebDriver
) -> WebDriver:
    """Close current tab of Browser;

    Parameters
    ----------
    driver : WebDriver
        Selenium WebDriver

    Returns
    -------
    WebDriver
        Selenium WebDriver
    """
    driver.close()
    return driver


def open_tab(
    driver: WebDriver, 
    url: str=''
) -> WebDriver:
    """Open new tab using JavaScript.

    Parameters
    ----------
    driver : WebDriver
        Selenium WebDrive
    url : str
        Open tab in the desired url

    Returns
    -------
    WebDriver
        Selenium WebDriver
    """
    
    try:
        driver.execute_script(f'window.open("{url}","_blank");')
        go_to_tab(driver, -1)
    except:
        go_to_tab(driver, 0)
        driver.execute_script(f'window.open("{url}","_blank");')
        go_to_tab(driver, -1)
    return driver


def go_to_tab(
    driver: WebDriver, 
    tab_index: int=0
) -> WebDriver:
    """Go to tab by index

    Parameters
    ----------
    driver : WebDriver
        Selenium WebDriver
    tab_index : int, optional
        Index of tab, from 0 to number of opened tabs
        minus 1, by default 0

    Returns
    -------
    WebDriver
        Selenium WebDriver
    """
    
    driver.switch_to.window(driver.window_handles[tab_index])
    return driver


def go_to_page(
    driver: WebDriver, 
    url: str
) -> WebDriver:
    """Open page by url

    Parameters
    ----------
    driver : WebDriver
        Selenium WebDriver
    url : str
        The desired URL page.

    Returns
    -------
    WebDriver
        Selenium WebDriver
    """
    driver.get(url)
    return driver


def find_element(
    query: str, 
    driver: WebDriver=None, 
    reference_el: WebElement=None, 
    by: By=By.XPATH
) -> WebElement:
    """Find element on page.

    Parameters
    ----------
    query : str
        The query string, often XPATH query string
    driver : WebDriver, optional
        Selenium Driver, by default None
    reference_el : WebElement, optional
        The reference element which query will be started
        from, by default None
    by : By, optional
        Search by option, by default By.XPATH

    Returns
    -------
    WebElement
        If find some element, returns the result of 
        search, else returns **None**

    Raises
    ------
    Exception
        If "driver" and "reference_el" is **None**
    """
    
    result = None
    try:
        if reference_el is not None:
            result = reference_el.find_element(by, value=query)
        elif driver is not None:
            result = driver.find_element(by, value=query)
        else:
            raise Exception('Provide "driver" or "reference_el".')
    except Exception as exc:
        logging.warning(exc.msg)
        result = None
    return result


def find_elements(
    query: str, 
    driver: WebDriver=None, 
    reference_el: any=None, 
    by: By=By.XPATH
) -> List[WebElement]:
    """Find elements on page.

    Parameters
    ----------
    query : str
        The query string, often XPATH query string
    driver : WebDriver, optional
        Selenium Driver, by default None
    reference_el : WebElement, optional
        The reference element which query will be started
        from, by default None
    by : By, optional
        Search by option, by default By.XPATH

    Returns
    -------
    List[WebElement]
        If find elements, returns the result of 
        search, else returns **None**

    Raises
    ------
    Exception
        If "driver" and "reference_el" is **None**
    """
    
    result = None
    try:
        if reference_el is not None:
            result = reference_el.find_elements(by, value=query)
        elif driver is not None:
            result = driver.find_elements(by, value=query)
        else:
            raise Exception('Provide "driver" or "reference_el".')
    except Exception as exc:
        logging.warning(exc.msg)
        result = None
    return result


def perform_click(
    element: WebElement,
    driver: WebDriver=None,
    js: bool=False
) -> None:
    """Perform Selenium click or Javascript click

    Parameters
    ----------
    element : WebElement
        The element which will be clicked
    driver : WebDriver, optional
        Selenium Driver, by default None
    js : bool, optional
        Perform click using Javascript, by default False
        
    Returns
    -------
    None
        **None** is returned
        
    """
    
    if js:
        assert driver is not None, \
            'When "js" == True, you need provide "driver"'
    if js:
        driver.execute_script("arguments[0].click();", element)
    else:
        element.click()


def click(
    element: WebElement, 
    driver: WebDriver=None, 
    js: bool=True, 
    retry: int=5, 
    js_when_exaust: bool=True
) -> None:
    """Click action with error handler

    Parameters
    ----------
    element : WebElement
        The element which will be clicked
    driver : WebDriver, optional
        Selenium Driver, by default None
    js : bool, optional
        Perform click using Javascript, by default True
    retry : int, optional
        Number of times the action will be attempted, by default 5
    js_when_exaust : bool, optional
        When click is not being performed by 
        Javascript ("js" == False), if True, after total number of 
        attempts, the click will try to be performed
        by Javascript, by default True

    Raises
    ------
    Exception
        When click action can not be sucessfuly executed.
    """
        
    retry_count = 0
    while retry_count < retry:
        try:
            perform_click(element, driver, js)
            retry_count = float('inf')
        except Exception as exc:
            retry_count = retry_count + 1
            msg = exc
            logging.debug(f'retry count: {retry_count}')
            logging.debug(msg)
    if retry_count != float('inf'):
        if js_when_exaust and driver is not None:
            perform_click(element, driver, js=True)
        else:
            raise Exception(msg)


def set_input_range_value(
    input_slide_element: WebElement, 
    driver: WebDriver, 
    value: int
) -> WebDriver:
    """Change value of slide input

    Parameters
    ----------
    input_slide_element : WebElement
        Slide input element
    driver : WebDriver
        Selenium WebDriver
    value : int
        Desired value to be placed on input

    Returns
    -------
    WebDriver
        Selenium WebDriver
    """
    curr_val = int(input_slide_element.get_attribute('value'))
    is_right_key = value > curr_val
    if is_right_key:        
        max_val = int(input_slide_element.get_attribute('max'))
        max_val = max(value, max_val)
        for i in range(max_val - curr_val):
            input_slide_element.send_keys(Keys.RIGHT)
    else:
        min_val = int(input_slide_element.get_attribute('min'))
        min_val = min(value, min_val)
        for i in range(curr_val, min_val, -1):
            input_slide_element.send_keys(Keys.LEFT)
    return driver


def get_el_on_el_list(xpath):
    """get xpath value that represent list, yield first el and increase
    by 1 iterator num"""
    pass


""" 
This module was created to handle DataFrames and returns 
pandas DataFrame or pandas Series.

"""


from drtools.utils import list_ops
from typing import List, Union, Tuple
from pandas import DataFrame, Series, cut
from scipy import stats
import numpy as np
import math
from copy import deepcopy
import pandas as pd


ByPercentage = 'by-percentage'
ByCount = 'by-count'
Lower = 'lower'
Upper = 'upper'


def combine_variables(
    dataframe: DataFrame,
    varibles: List[str],
    sep: str='|'
) -> DataFrame:
    """Combine categorical variables of DataFrame in order to analyse 
    impact of combined variables on response.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame
    varibles : List[str]
        List of categorical variables that will be combined
    sep : str, optional
        Separation character when combine variables, 
        by default '|'

    Returns
    -------
    DataFrame
        The DataFrame with new colum containing all combinations 
        of desired variables.
    """
    
    df = dataframe.copy()
    title = sep.join(varibles)
    def get_all_values(x):
        text = ''
        variables_len = len(varibles)
        for idx, var in enumerate(varibles):
            text += str(x[var])
            if idx != variables_len - 1:
                text += str(sep)
        return text
    df[title] = df.apply(get_all_values, axis=1)
    return df


def remove_boundary_values(
    dataframe: DataFrame,
    by: Union[ByPercentage, ByCount]=ByPercentage,
    value: float=0.05,
    boundary: Union[Lower, Upper]=Upper
) -> DataFrame:
    """Remove values of top part of the DataFrame

    Parameters
    ----------
    df : DataFrame
        The DataFrame
    by : Union[ByPercentage, ByCount], optional
        Rule that will be applied to remove top
        values, by default ByPercentage
    value : float, optional
        The value of the rule, by default 0.05
    boundary : Union[Lower, Upper], optional
        Extremity from where values 
        will be removed, by default 0.05

    Returns
    -------
    DataFrame
        The DataFrame after removing selected values of the 
        top part of DataFrame

    Raises
    ------
    ValueError
        If **by** option is invalid.
    """
    
    methods = {
        Lower: {
            ByPercentage: math.floor(value * len(dataframe)),
            ByCount: value,
            'df_filter': lambda x: dataframe.iloc[x:]
        },
        Upper: {
            ByPercentage: math.ceil((1-value) * len(dataframe)),
            ByCount: value,
            'df_filter': lambda x: dataframe.iloc[:x]
        }
    }    
    limit = methods[boundary][by]
    df_copy = methods[boundary]['df_filter'](limit)
    return df_copy


def remove_outliers(
    df: DataFrame,
    col_name: str,
    zscore_limit: int=3,
    fillna: any=None
) -> DataFrame:
    """Remove outliers using zscore rule
    
    Only will be keeped values with zscore lower than **zscore_limit**

    Parameters
    ----------
    df : DataFrame
        The DataFrame
    col_name : str
        The column name where outliers will be removed
    zscore_limit : int, optional
        Z-Score limit, by default 3
    fillna : any, optional
        Value to fill na before apply the zscore rule, by default None

    Returns
    -------
    DataFrame
        The DataFrame after removing the outliers
    """
    df_copy = df.copy()
    if fillna is not None:
        df_copy[col_name] = df_copy[col_name].fillna(fillna)
    df_copy = df[
        np.abs(stats.zscore(df_copy[col_name])) < zscore_limit
    ]
    return df_copy


ColumnInfo = dict({
    str: float
})


def keep_categories(
    df: DataFrame,
    col_name: str,
    keep_categories: List[str]=[]
) -> DataFrame:
    """Keep only selected categories from column of DataFrame.

    Parameters
    ----------
    df : DataFrame
        The DataFrame
    col_name : str
        The column name where categories will be keeped
    keep_categories : List[str], optional
        List of categories that will be keeped, by default []

    Returns
    -------
    DataFrame
        The DataFrame after remove unselected categories and keeped
        selected categories
    """
    df_copy = df[col_name].isin(
        keep_categories
    )
    return df[df_copy]


def join_categories(
    dataframe: DataFrame,
    col_name: str,
    categories: List[str]=[], 
    to_category: str='others'
) -> DataFrame:
    """Join categories and fill these values with desired value.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame
    col_name : str
        The column name where categories will be joined.
    categories : List[str], optional
        List of categories will be joined, by default []
    to_category : str, optional
        New value for selected categories, by default 'others'

    Returns
    -------
    DataFrame
        The DataFrame after join the categories
    """
    df = dataframe.copy()
    if len(categories) == 0:
        return df
    condition = df[col_name].isin(categories)
    df.loc[condition, col_name] = to_category
    return df


def keep_categories_and_join_remaining(
    dataframe: DataFrame,
    column: str,
    categories: List[str]=[], 
    to_category: str='others'
) -> DataFrame:
    """Join categories and fill these values with desired value.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame
    column : str
        The column name where categories will be joined.
    categories : List[str], optional
        List of categories will be joined, by default []
    to_category : str, optional
        New value for selected categories, by default 'others'

    Returns
    -------
    DataFrame
        The DataFrame after join the categories
    """
    df = dataframe.copy()
    if len(categories) == 0:
        return df
    df[column] = np.where(
        pd.isnull(df[column]), 
        df[column], 
        df[column].astype(str)
    ) 
    join_cat = list(df[column].dropna().unique())
    join_cat = list_ops(join_cat, categories)
    df = join_categories(df, column, join_cat, to_category)
    return df


def drop_categories(
    dataframe: DataFrame,
    col_name: str,
    categories: List[str],
) -> DataFrame:
    """Drop categories of column

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame
    col_name : str
        The column name where categories will be dropped
    categories : List[str]
        List of categories that will be dropped

    Returns
    -------
    DataFrame
        The DataFrame after drop the categories
    """
    condition = ~dataframe[col_name].isin(
        categories
    )
    return dataframe[condition]


def get_categories_ge(
    df: DataFrame,
    col_name: str,
    by: Union[ByPercentage, ByCount]=ByPercentage,
    value: float=0.01,
    ignore_categories: List[str]=[]
) -> List[str]:
    """Get name of categories that contains percentage or counts 
    of total samples greater or equal than value.

    Parameters
    ----------
    df : DataFrame
        DataFrame where categoreis will be found
    col_name : str
        Column name where categories will be found
    by : Union[ByPercentage, ByCount], optional
        Rule that will be applied to get categories based
        on percentage of total or count, by default ByPercentage
    value : float, optional
        The value of the rule, by default 0.01
    ignore_categories : List[str], optional
        Ignore this categories when apply percentage 
        filter, by default []

    Returns
    -------
    List[str]
        The list of categories with at least desired percentage of total
    """
    resp = []
    v = df[col_name].value_counts()
    col_len = df[col_name].notna().sum()
    
    for category, category_count_value in zip(v.index, v.values):
        if category in ignore_categories:
            continue
        
        if by == ByPercentage:
            compare_value = category_count_value / col_len
        elif by == ByCount:
            compare_value = category_count_value
        else:
            raise ValueError('Invalid "by" option.')       
        
        if compare_value >= value:
            resp.append(category)
    return resp


def get_labels_from_bin_interval(
    bins: List[float],
) -> List[str]:
    """Get bins labels from bins interval of values

    Parameters
    ----------
    bins : List[float]
        List containing bins boundary

    Returns
    -------
    List[str]
        The list with labels for each interval
    """
    return [f'{x}_{y}' for x, y in zip(bins[:-1], bins[1:])]


def prepare_bins(
    bins: Union[int, List[float]],
    smallest_value: float=None,
    bigger_value: float=None,
    include_upper: bool=False
) -> Tuple[List[float], List[float], List[str]]:
    """Generate values of bins, middle points and labels for the bins
    
    This function will generate points which represents the x axis
    for the bins, the middle points of the intervals of the bins and
    the labels for each bin.

    Parameters
    ----------
    bins : Union[int, List[float]]
        The desired number of bins that will be generated
        or the bins interval
    smallest_value : float, optional
        Smallest value from the interval, by default None.
    bigger_value : float, optional
        Bigger value from the interval, by default None.
    include_upper : bool, optional
        Whether the last interval should 
        be right-inclusive or not, by default False

    Returns
    -------
    Tuple[List[float], List[float], List[str]]
        Returns the values which represents the intervals 
        from the x axis, the middle points of these intervals and
        the labels for the bins interval.
    """
    values = deepcopy(bins) \
        if hasattr(bins, '__iter__') \
        else np.linspace(smallest_value, bigger_value, bins + 1)
    if include_upper:
        upper_boundary = round(values[-1] * (1 + 0.001), 4) \
            if values[-1] != 0 \
            else 1e-4
        values[-1] = upper_boundary
    middle_points = [(x + y) / 2 for x, y in zip(values[:-1], values[1:])]
    labels = get_labels_from_bin_interval(values)
    labels = [
        f'[{label})'.replace('_', ', ')
        for label in labels
    ]
    return values, middle_points, labels
    

def binning_numerical_variable(
    dataframe: DataFrame,
    col_name: str,
    bins: Union[int, List[float]],
    binning_column_prefix: str='binning',
    include_upper: bool=False,
    labels: List[str]=None
) -> DataFrame:
    """Binning numerical variable.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame from where numerical variable will be found.
    col_name : str
        The numerical column name.
    bins : Union[int, List[float]]
        Desired number of bins or interval
        of bins
    binning_column_prefix : str, optional
        Prefix of new columns that will be generated
        with the binning of numerical variable, by default 'binning'
    include_upper : bool
        Whether the last interval should 
        be right-inclusive or not, by default False
    labels : List[str]
        Labels for the bins interval, by default **None**

    Returns
    -------
    DataFrame
        The DataFrame with new column containing the binned values
        for the numerical variable
    """
    
    df = dataframe.copy()
        
    sorted_values = df[col_name] \
        .dropna() \
        .sort_values() \
        .values
    
    smallest_value = sorted_values[0]
    bigger_value = sorted_values[-1]
    values, middle_points, _labels = prepare_bins(
        bins,
        smallest_value=smallest_value,
        bigger_value=bigger_value,
        include_upper=include_upper
    )
    df[f'{binning_column_prefix}_{col_name}'] = cut(
        df[col_name], 
        bins=values,
        labels=labels or _labels,
        right=False
    )
    
    return df


def join_categories_lt(
    dataframe: DataFrame,
    col_name: str,
    by: Union[ByPercentage, ByCount]=ByPercentage,
    value: float=0.01,
    to_category_name: str='others',
) -> DataFrame:
    """Fill value with 'to_category_name' of categories 
    which has count lower than 'value' in count or percentage

    Parameters
    ----------
    dataframe : DataFrame
        DataFrame
    col_name : str
        Name of categorical column
    by : Union[ByPercentage, ByCount], optional
        Rule that will be applied to join categories based
        on percentage of total or count, by default ByPercentage
    value : float, optional
        The value of the rule, by default 0.01
    to_category_name : str, optional
        Value to fill categories, by default 'others'

    Returns
    -------
    DataFrame
        DataFrame after applied changes. 
    """
    df = dataframe.copy()
    keep_categories = get_categories_ge(
        df, 
        col_name, 
        by=by,
        value=value,
    )
    all_categories = list(df[col_name].value_counts().index)
    join_categories_list = list_ops(
        all_categories, 
        keep_categories
    )
    df = join_categories(
        df, 
        col_name, 
        categories=join_categories_list,
        to_category=to_category_name
    )
    return df


def keep_categories_ge(
    df: DataFrame,
    col_name: str,
    by: Union[ByPercentage, ByCount]=ByPercentage,
    value: float=0.01,
    force_remove: List[str]=[]
) -> DataFrame:
    """Keep categories of 'col_name' which contains 
    percentage or counts greater or equal than 'value'.

    Parameters
    ----------
    df : DataFrame
        Pandas DataFrame
    col_name : str
        Column name to keep categories
    by : Union[ByPercentage, ByCount], optional
        Rule that will be applied to keep categories based
        on percentage of total or count, by default ByPercentage
    value : float, optional
        Keep categories which has at least
        this percentage or count of total 
        samples, by default 0.01
    force_remove : List[str], optional
        Remove columns from this list even if this columns
        doesn't match the percentage condition, by default []

    Returns
    -------
    DataFrame
        DataFrame with only desired categories
    """
    
    keep_categories = get_categories_ge(
        df, 
        col_name, 
        by=by,
        value=value,
    )
    keep_categories = list_ops(
        keep_categories, 
        force_remove
    )
    df_copy = keep_categories(
        df, 
        col_name, 
        keep_categories=keep_categories
    )
    return df_copy


def na_percentage_of_column(
    df: DataFrame,
    col_name: str,
) -> float:
    """Number representing NA percentage from column of DataFrame 

    Parameters
    ----------
    df : DataFrame
        The DataFrame
    col_name : str
        Columns name to verify nullity

    Returns
    -------
    float
        The percentage of na for the columns passed as 'col_name'
        argument
    """
    return df[col_name].isna().sum() / len(df[col_name])


def percentage_of_most_common_value_by_column(
    df: DataFrame,
    ignore_columns: List[str]=[],
    notna: bool=False
) -> ColumnInfo:
    """Percentage of the most common value for each column.

    Parameters
    ----------
    df : DataFrame
        Pandas DataFrame
    ignore_columns : List[str], optional
        Ignore columns when get percentage, by default []
    notna : bool, optional
        If True, percentage will be based on not na values,
        if False, percentage will be calculated considering na
        values, by default False

    Returns
    -------
    ColumnInfo
        Dict containing information about percentage of most common
        value from each column.
    """
    
    resp = {}
    for col in df.columns:
        if col in ignore_columns:
            continue
        
        data_series = df[col].dropna() \
            if notna \
            else df[col].fillna('nan')
        
        v = data_series.value_counts()
        num_of_occ = v.sort_values().values
        num_of_occ = num_of_occ[-1]
        total_len = len(data_series)
        value_percentage = num_of_occ / total_len
        
        resp[col] = round(value_percentage * 100, 2)
    return resp


def na_percentage_info(
    df: DataFrame,
    ignore_columns: List[str]=[]
) -> ColumnInfo:
    """Get percentage of null of all columns.

    Parameters
    ----------
    df : DataFrame
        Pandas DataFrame.
    ignore_columns : List, optional
        Ignore these columns when check DataFrame, by default []

    Returns
    -------
    ColumnInfo
        Information of null percentage about each column
    """
    
    resp = {}
    for col in df.columns:
        if col not in ignore_columns:
            resp[col] = round(
                na_percentage_of_column(df, col) * 100, 
                2
            )
    return resp


def get_column_names(
    df: DataFrame,
    contains: List[str]=[]
) -> List[str]:
    """Get name of columns that contains some of substrings in 'contains'

    Parameters
    ----------
    df : DataFrame
        DataFrame from which columns will be generated
    contains : List[str], optional
        List of substring to filter DataFrame columns, by default []

    Returns
    -------
    List[str]
        The list of column names which has some of the substrings
        in 'contains'
    """
    
    names = []
    if len(contains) == 0:
        contains = ['']
    for col in df.columns:
        for col_name in contains:
            if col_name in col:
                names.append(col)
                break
    return names


def drop_df_columns_gt_constancy(
    dataframe: DataFrame,
    constacy_level: float=0.75,
    verbose: bool=True,
    print_removed_columns: bool=False,
) -> DataFrame:
    """Drop columns based on variable constancy.
    
    Drop columns that has some class or value with count percentage 
    bigger than constancy_level.

    Parameters
    ----------
    dataframe : DataFrame
        DataFrame that columns will be dropped
    constacy_level : float, optional
        Columns that has some value with percentage of counts 
        above constancy level will be dropped, must be
        between 0 and 1, by default 0.8
    verbose : bool, optional
        If True print extra information, by default True
    print_removed_columns : bool, optional
        If True print removed columns, by default False

    Returns
    -------
    DataFrame
        DataFrame after drop columns
    """
    df = dataframe.copy()
    drop_columns = []
    for col in df.columns:
        v = df[col].value_counts()
        num_of_occ = v.sort_values().values[-1]
        total_len = len(df[col])
        variable_constancy_level = num_of_occ / total_len
        if variable_constancy_level >= constacy_level:
            drop_columns.append(col)
    df.drop(columns=drop_columns, axis=1, inplace=True)
    if verbose:
        len_initial = len(dataframe.columns)
        len_final = len(df.columns)
        print(f'From {len_initial} to {len_final}. Remove: {len_initial - len_final}')
    if print_removed_columns:
        removed = [col for col in df.columns if col not in df.columns]
        print(removed)
    return df


def merge_columns(
    dataframe: DataFrame,
    cols: List[str],
) -> Series:
    """Create pandas Series containing the result of 
    combination of some list of columns.
    
    The resulting Series will contain the first filled value 
    from a column passed on variable `cols`.

    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame.
    cols : List[str]
        List of column names.

    Returns
    -------
    Series
        The resulting Series.
    """
    
    df = dataframe[cols].copy()
    series = dataframe[cols[0]]
    
    for col in cols[1:]:
        series = series.fillna(df[col])
    
    return series


def reduce_mem_usage(
    dataframe: DataFrame, 
    verbose=True
):
    """Reduce memory usage from DataFrame
    
    Parameters
    ----------
    dataframe : DataFrame
        The DataFrame.
    verbose : bool, optional
        If True, will print reduction information, 
        by default True.

    Returns
    -------
    DataFrame
        The resulting DataFrame after reduce memory usage.
    """    
    
    df = dataframe.copy()
    numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
    start_mem = df.memory_usage().sum() / 1024**2
    
    for col in df.columns:
        col_type = df[col].dtypes
        if col_type in numerics:
            c_min = df[col].min()
            c_max = df[col].max()
            if str(col_type)[:3] == 'int':
                if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                    df[col] = df[col].astype(np.int8)
                elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                    df[col] = df[col].astype(np.int16)
                elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                    df[col] = df[col].astype(np.int32)
                elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                    df[col] = df[col].astype(np.int64)
            else:
                if c_min > np.finfo(np.float16).min and c_max < np.finfo(np.float16).max:
                    df[col] = df[col].astype(np.float16)
                elif c_min > np.finfo(np.float32).min and c_max < np.finfo(np.float32).max:
                    df[col] = df[col].astype(np.float32)
                else:
                    df[col] = df[col].astype(np.float64)

    end_mem = df.memory_usage().sum() / 1024**2
    
    if verbose:
        print('Memory usage BEFORE optimization is: {:.2f} MB'.format(start_mem))
        print('Memory usage AFTER optimization is: {:.2f} MB'.format(end_mem))
        print('Decreased by {:.1f}%'.format(100 * (start_mem - end_mem) / start_mem))

    return df


""" 
This module was created to generate 
useful Decorators.

"""

def start_end_log(method):
    """Decorator to print name of execution method before and 
    after method execution.

    Parameters
    ----------
    method : str
        The method name that will be executed.
    """
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            if self.LOGGER is not None:
                self.LOGGER.debug(f'Executing {method}()...')
                
            # execution
            response = f(self, *args, **kwargs)
            
            if self.LOGGER is not None:
                self.LOGGER.debug(f'Executing {method}()... Done!')
                
            return response
        wrapper.__doc__ = f.__doc__
        return wrapper
    return decorator